//===============================================
//            GoGo Board PIC Code
//
//===============================================
//
//            Version 1.60.ms1
//              Mar 9, 2003
//
//
//	  Works only with GoGo board 1.6 and higher
//     See ChangeLog.txt for version history.
//
//
//By Arnan Sipitakiat
//arnans@media.mit.edu
//
//Future of Learning Group
//MIT Media Laboratory
//(C) 2002 MIT Media Laboratory
//

#case

// Define which MCU you are using here.
// This code has been used with the 16F871, 16F877, and 16F77 chips.
// It should work with the other chips in the same family as well.
#include <16F877.H>
#DEFINE PIC16F877



#device ADC=10 *=16

#IFDEF PIC16F77
// the F77 doesn't support NOLVP (No Low Volt Prog)
#fuses HS,NOWDT,NOPROTECT, NOBROWNOUT
#ELSE
// Without the NOLVP (Low Volt Prog) flag, I think pins RB3,RB6,RB7
// will become unusable. One side effect is
// the green LED will not function (it uses RB3).
#fuses HS,NOWDT,NOPROTECT, NOBROWNOUT, NOLVP
#ENDIF




#use DELAY(clock=20000000)

#include <GOGO16.H>
#include <ser_87x.h>

#include <ser_87x.c>


#use fast_io(A)
#use fast_io(B)
//#use fast_io(C)  // somehow this doesn't work. Compiler bug?
#use fast_io(D)
#use fast_io(E)



// Define the RF pins
//#use rs232(baud=2400, /*xmit=RF_XMIT,*/ rcv=PIN_B3, invert)
//#define PIC_TRIS_B  0b00001100


//#use rs232(baud=2400, xmit=RF_XMIT, rcv=RF_RCV, invert)


#define defaultPort		0
#define channelSwitchDelay	1	// delay time in ms after switching adc channels


/////////////////////////////////////////////////////////////////
//  Function Declaration
/////////////////////////////////////////////////////////////////
void setHigh(int Pin);
void setLow(int Pin);
int  readPin(int Pin);

short getBit(int InByte, int BitNo);
void setBit(int *InByte, int BitNo);
void clearBit(int *InByte, int BitNo);

void Ping(int Param);
void TalkToMotor(int MotorBits);
#inline void MotorControl(int MotorCmd);
#inline void SetMotorPower(int Power);
void MotorON(int MotorNo);
void MotorOFF(int MotorNo);
void MotorRD(int MotorNo);
void MotorThisWay(int MotorNo);
void MotorThatWay(int MotorNo);
void MotorCoast(int MotorNo);
void SetBurstMode(int SensorBits);
void DoSensorStuff();
#inline  void ReadSensor(int Target, int readMode);
long getSensorVal();
void switchAdcChannel(int channelNo);

void ProcessInput();
//int  process_input();
//void Add_to_CMD(int InByte);
//void EndCMD();
//byte get_cmd_buff(int *cmd_ptr);
void init_variables();
void intro ();
void Halt();
void initBoard();


/////////////////////////////////////////////////////////////////
//  Global Variables
//
//  Please becarfule with global variables.
//  I have experienced some strange problems
//  whith them, especially when many globals are
//  used. I think there's a bug in the compiler
//  that messes up globals that are passed as
//  a function parameter. See description in
//  motorControl() for more info.
/////////////////////////////////////////////////////////////////

int  MotorPins [18]={  MTR1_EN, MTR1_CW, MTR1_CC,
MTR2_EN, MTR2_CW, MTR2_CC,
MTR3_EN, MTR3_CW, MTR3_CC,
MTR4_EN, MTR4_CW, MTR4_CC,
MTR5_EN, MTR5_CW, MTR5_CC,
MTR6_EN, MTR6_CW, MTR6_CC };

int CMD_STATE;
short PASS_THROUGH;
int gbl_cur_cmd, gbl_cur_param, gbl_cur_ext, gbl_cur_ext_byte;
int gblBurstModeBits;

int gblCurSensorChannel;
int gblActiveMotors;
int gblMotorDir=0;
int gblMotorONOFF = 0;
int gblMtrDuty[6] = {0xff,0xff,0xff,0xff,0xff,0xff};  // Motor PWM bits
int gblDutyCount = 0;


// New in 1.4 MS2
// These arrays store sensor data in different modes
unsigned int gblSSMax[8]={0};
unsigned int gblSSMin[8];       // it gets initialized to 255 in init_variables()
                                // gblSSMin[8]={255} doesn't work for some reason.
//unsigned int gblSSMean[8]={255};






//int bglIntCount;


//#define TIMER_FREQ	3		// int per second = 76 / TIMER_FREQ



#int_rtcc                          // This function is called every time
clock_isr() {                      // the RTCC (timer0) overflows (255->0).
                                   // For this program this is apx 76 times
                                   // per second.
int i;

for (i=0;i<6;i++)
{	if (getBit(gblMotorONOFF,i) == ON)
	{ if ((gblMtrDuty[i] >> gblDutyCount) & 1)
		{
			// I hard coded the following command for better performance.
			//setLow(MotorPins[3*i]);

			switch (i) {
				case 0: rb0 = 1; break;
				case 1: portd5 = 1; break;
				case 2: rb3 = 1; break;
				case 3: portc4 = 1; break;
				case 4: rb6 = 1; break;
				case 5: portd0 = 1; break;
			}
		} else
		{	// I hard coded the following command for better performance.
			//setHigh(MotorPins[3*i]);
			switch (i) {
				case 0: rb0 = 0; break;
				case 1: portd5 = 0; break;
				case 2: rb3 = 0; break;
				case 3: portc4 = 0; break;
				case 4: rb6 = 0; break;
				case 5: portd0 = 0; break;
			}

		}
	}
}
gblDutyCount++;
gblDutyCount &= 0b000111;  // limit the max value to 31

}




//////////////////////////////////////////////////
//   In the following few functions ..
//
//    (Pin/8) is the port number (5=PortA, 6=B, ...)
//    (Pin&7) is the pin number  (0-7)
//
//    So,
//
//    *(Pin/8) is a pointer to the port
//    1<<(Pin&7) is a bit mask used to set/get a bit
//
//    Consult the PIC-C FAQ for more info.

void setLow(int Pin)
{
   *(Pin/8) &= (~(1<<(Pin&7)));
}

void setHigh(int Pin)
{
   *(Pin/8) |= (1<<(Pin&7));

}


int readPin(int Pin)
{
   return (*(Pin/8) & (1<<(Pin&7))) ;

}



//////////////////////////////////////////////////
short getBit(int InByte, int BitNo)
{  return ((InByte >> BitNo) & 1);
}

void setBit(int *InByte, int BitNo)
{  *InByte |= (1<<BitNo);
}

void clearBit(int *InByte, int BitNo)
{  *InByte &= ~(1<<BitNo);
}

/////////////////////////////////////////////////


void Ping(int Param)
{
   RX_LED_ON;
	if (Param != 0)
	{
		PASS_THROUGH = TRUE;
	} else
	{	PASS_THROUGH = FALSE;
	}
	printf(ser_char,"%c%c%c",ReplyHeader1, ReplyHeader2, ACK_BYTE);  // send acknowledge byte
   RX_LED_OFF;

}


/////////////////////////////////////////////////////////////////////////
//
//   M O T O R   C O N T R O L
//
/////////////////////////////////////////////////////////////////////////

void TalkToMotor(int MotorBits)
{
   RX_LED_ON;
	// Each bit represents one motor. i.e 0b00010010 means motor 2 and 5 are active
	gblActiveMotors = MotorBits;

	printf(ser_char,"%c%c%c", ReplyHeader1, ReplyHeader2, ACK_BYTE);  // send acknowledge byte
   RX_LED_OFF;
}


// While adding code to GoGo1.4 ms2, I had a problem with parameters passed to
// procedures (like this one) not receiving the correct values.  Making it an
// inline procedure fixed it. I think this is a compiler bug.
//
// Without #inline, MotorCmd would get strange values. I believe this problem
// startes to happen when we use lots of global variables.
#inline void MotorControl(int MotorCmd)
{
	int i,j;

   RX_LED_ON;

	for (i=0;i<6;i++)
	{
		if ((gblActiveMotors >> i) & 1 )
		{

//			printf(ser_char,"%c",i);

			switch (MotorCmd)
//			switch (gbl_cur_param)

			{
			case 0:	MotorON(i);
					break;
			case 1: MotorOFF(i);
					break;
			case 2: MotorRD(i);
					break;
			case 3: MotorThisWay(i);
					break;
			case 4: MotorThatWay(i);
					break;
			case 5: MotorCoast(i);
					break;
			}

		}
	}


	printf(ser_char,"%c%c%c", ReplyHeader1, ReplyHeader2, ACK_BYTE);  // send acknowledge byte

   RX_LED_OFF;
}

/////////////////////////////////////////////////
// Refer to the problem explained in MotorControl() if
// this function suddenly missbehave
#inline void SetMotorPower(int Power)
{
	int i;

//	Power = 8 - Power;    // 0 = full power, 8 = off

   RX_LED_ON;

	for (i=0;i<6;i++)
	{
		if ((gblActiveMotors >> i) & 1 )
		{

			switch (Power) {
//				case 0: Power = 0x00; break;  // commented out because values don't change anyway
//				case 1: Power = 0x01; break;
				case 2: Power = 0x11; break;
				case 3: Power = 0xAA; break;
				case 4: Power = 0xEE; break;
				case 5: Power = 0xFE; break;
				case 6:
				case 7: Power = 0xFF; break;
			}
			gblMtrDuty[i] = Power;

		}
	}


	printf(ser_char,"%c%c%c", ReplyHeader1, ReplyHeader2, ACK_BYTE);  // send acknowledge byte

   RX_LED_OFF;

}


void MotorON(int MotorNo)
{	int Index;
	int MtrEnable, MtrCC, MtrCW;

	Index = MotorNo * 3;

	MtrEnable 	= MotorPins[Index];
	MtrCW 		= MotorPins[Index + 1];
	MtrCC 		= MotorPins[Index + 2];


	if (getBit(gblMotorDir,MotorNo))
		{ 	setLow(MtrCC); setHigh(MtrCW);	}
	else
		{	setHigh(MtrCC); setLow(MtrCW);	}
	setHigh(MtrEnable);


setBit(&gblMotorONOFF,MotorNo);
}



void MotorOFF(int MotorNo)
{	int Index;
	int MtrEnable, MtrCC, MtrCW;

	Index = MotorNo * 3;

	MtrEnable 	= MotorPins[Index];
	MtrCW 		= MotorPins[Index + 1];
	MtrCC 		= MotorPins[Index + 2];

   if (readPin(MtrCW))
      setBit(&gblMotorDir,MotorNo);
   else
      clearBit(&gblMotorDir, MotorNo);

  	clearBit(&gblMotorONOFF,MotorNo);

	setHigh(MtrEnable);     // this ensures the motor breaks. Sometimes the timer_isr turns off the enable pin, which turns 'stop' into 'coast'
	setHigh(MtrCC);
	setHigh(MtrCW);


}



void MotorRD(int MotorNo)
{	int Index;
	int MtrCC, MtrCW;

	Index = MotorNo * 3;

	MtrCW 		= MotorPins[Index + 1];
	MtrCC 		= MotorPins[Index + 2];


	if (getBit(gblMotorONOFF,MotorNo) == ON)
	{

		if (getBit(gblMotorDir,MotorNo))
		{	setLow(MtrCW);
			setHigh(MtrCC);
         clearBit(&gblMotorDir,MotorNo);
		} else
		{	setHigh(MtrCW);
			setLow(MtrCC);
         setBit(&gblMotorDir,MotorNo);
		}
	}


}


///////////////////////////////////////////////////////////////////


void MotorThisWay(int MotorNo)
{	int Index;
	int MtrCC, MtrCW;

	Index = MotorNo * 3;

	MtrCW 		= MotorPins[Index + 1];
	MtrCC 		= MotorPins[Index + 2];

	setBit(&gblMotorDir,MotorNo);


	if (getBit(gblMotorONOFF,MotorNo) == ON)
	{
		setLow(MtrCC);
		setHigh(MtrCW);
	}

}


void MotorThatWay(int MotorNo)
{	int Index;
	int MtrCC, MtrCW;

	Index = MotorNo * 3;

	MtrCW 		= MotorPins[Index + 1];
	MtrCC 		= MotorPins[Index + 2];

	clearBit(&gblMotorDir,MotorNo);

	if (getBit(gblMotorONOFF,MotorNo) == ON)
	{
		setLow(MtrCW);
		setHigh(MtrCC);
	}

}


void MotorCoast(int MotorNo)
{
  	clearBit(&gblMotorONOFF,MotorNo);
	setLow(MotorPins[MotorNo * 3]);
}

/////////////////////////////////////////////////////////////////////////
//
//   P R O C E S S   S E N S O R    W H I L E    I D L E
//
/////////////////////////////////////////////////////////////////////////



void DoSensorStuff()
{
	int i;
	long SensorVal;
   unsigned int SensorVal8Bit;
//   int hold;

		for (i=0;i<8;i++)
		{
				if (gblCurSensorChannel != i)
				{	switchAdcChannel(i);
					gblCurSensorChannel=i;
				}
				SensorVal=getSensorVal();

            // if this sensor is in Burst mode -> send value to computer
            if ((gblBurstModeBits>>i) & 1)
            {  GREEN_LED_ON;
               printf(ser_char,"%c%c%c", 0x0c, (i << 5) | (SensorVal >> 8), SensorVal & 0xff);
               GREEN_LED_OFF;
            }


            // reduce the sensor resolution from 10 to 8 bits,
            // as MAX, MIN, MEAN values are 8 bits (to save memory)

            SensorVal8Bit = (int) (SensorVal >> 2);

            /// If this is a new MAX
            if (SensorVal8Bit > gblSSMax[i])
               gblSSMax[i] = SensorVal8Bit;
            /// If this is a new MIN
            if (SensorVal8Bit < gblSSMin[i])
               gblSSMin[i] = SensorVal8Bit;

            /// Recalculate Mean.
            /// This algorithm is just an approximation of Mean.
            /// It should give sensor readings emulating
            /// the situation where there is a capacitor connected accross
            /// the sensor. That is, it slows down fluctuation.
/*
            hold =  gblSSMean[i];
            hold -=  hold >> 3;
            gblSSMean[i] = hold + (SensorVal8Bit >> 3);
*/		}


}


/////////////////////////////////////////////////////////////////////////
//
//   S E N S O R   C O N T R O L
//
/////////////////////////////////////////////////////////////////////////


void SetBurstMode(int SensorBits)
{
   RX_LED_ON;
	gblBurstModeBits = SensorBits;

	printf(ser_char,"%c%c%c", ReplyHeader1, ReplyHeader2, ACK_BYTE);
   RX_LED_OFF;
}



/////////////////////////////////////////////////
// See motorControl() for explanation why I used
// #inline here.
#inline void ReadSensor(int Target, int readMode)
{
long SensorVal;

RX_LED_ON;
printf(ser_char,"%c%c", ReplyHeader1, ReplyHeader2);

switch (readMode) {

   case MAX_READ:
      SensorVal = (long)gblSSMax[Target] << 2;
      gblSSMax[Target] = 0;
      break;

   case MIN_READ:
      SensorVal = (long)gblSSMin[Target] << 2;
      gblSSMin[Target] = 255;
      break;

   case NORMAL_READ:
   default:

      if (gblCurSensorChannel !=Target)
      {	switchAdcChannel(Target);
     	   gblCurSensorChannel=Target;
      }
      SensorVal=getSensorVal();

      break;



/*
   case MEAN_READ:
      SensorVal = (long)gblSSMean[Target] << 2;
      break;
*/
}

printf(ser_char,"%c%c", SensorVal >> 8, SensorVal & 0xff);

RX_LED_OFF;

}

/////////////////////////////////////////////////
long getSensorVal()
{
   long sensorVal;

   sensorVal=read_adc();
   //return (sensorVal >> 6);  // use this line if using PIC-C compiler version 2.x

   // the PIC 16F77 ADC is only 8 bits. So, we simulate 10 bits by shifting
   // left 2 bits. But why in the code we shif right 6 bits? This is because
   // somehow the 8 bit sensor readings are stored in the higher byte
   // of the long variable. So, we have to shift right 8 bits before shifting
   // left. So, 8-2 is 6.
   #IFDEF PIC16F77
   sensorVal >>= 6;
   #ENDIF



   return (sensorVal);        // this line works with PIC-C compiler version 3.x

}

/////////////////////////////////////////////////
void switchAdcChannel(int channelNo) {
set_adc_channel(channelNo);
delay_ms(channelSwitchDelay);   // wait for adc to stabilize
}




/////////////////////////////////////////////////////////////////////////
//
//   M I S C
//
/////////////////////////////////////////////////////////////////////////

void init_variables()
{  int i;
	gblBurstModeBits = 0;

	CMD_STATE = WAITING_FOR_FIRST_HEADER;
	PASS_THROUGH = FALSE;

   for (i=0;i<7;i++)
   {  gblSSMin[i] = 255;
   }

}


void intro ()
{
	RX_LED_ON; delay_ms(100); RX_LED_OFF;
	TX_LED_ON; delay_ms(100); TX_LED_OFF;
	RX_LED_ON; delay_ms(100); RX_LED_OFF;
}


void Halt()
{	while (1)
	{	TX_LED_ON;
		RX_LED_ON;
	}
}



void initBoard()
{
	int i,j;

	gblActiveMotors = 0;

	set_tris_a(PIC_TRIS_A);
	set_tris_b(PIC_TRIS_B);
	//TRISC = PIC_TRIS_C;
	set_tris_c(PIC_TRIS_C);
	set_tris_d(PIC_TRIS_D);
	set_tris_e(PIC_TRIS_E);

	// Init the analog ports
	setup_port_a( ALL_ANALOG );
	setup_adc( ADC_CLOCK_INTERNAL );

	// Init default sensor port
	gblCurSensorChannel=defaultPort;
	set_adc_channel( defaultPort );


	RX_LED_OFF;
	TX_LED_OFF;

	// Init all six motors' turn direction to CW
	for (i=0,j=0 ; i<6 ; i++,j=i*3)
	{	setLow(MotorPins[j]);       // Mtr Enable pin
		setHigh(MotorPins[j + 1]);	// Mtr CW pin
		setLow(MotorPins[j + 2]);	// Mtr CCW pin

	}


	// init the serial port
	ser_init();
//	open_rx_com();
//	pspmode = 0;



	   // setup the TIMER interrupt
//	   gblIntCount=TIMER_FREQ;
	   set_rtcc(0);
	   setup_counters( RTCC_INTERNAL, RTCC_DIV_32);
	   enable_interrupts(GLOBAL);
      enable_interrupts(INT_RTCC);
      enable_interrupts(INT_RDA);



   intro();


}










/////////////////////////////////////////////////////////////////////////
//
//   S E R I A L    D A T A    H A N D L E R
//
/////////////////////////////////////////////////////////////////////////


void ProcessInput()
{	byte InByte, buff_status;

	while ((buff_status = get_rx_buff(&InByte)) == SUCCESS)  // fetch character
	{
		   	switch (CMD_STATE) {
				case WAITING_FOR_FIRST_HEADER:
					 if (InByte == InHeader1) CMD_STATE = WAITING_FOR_SECOND_HEADER;
					 break;
				case WAITING_FOR_SECOND_HEADER:
					 if (InByte == InHeader2) CMD_STATE = WAITING_FOR_CMD_BYTE;
					 break;
				case WAITING_FOR_CMD_BYTE:
			  		gbl_cur_cmd 	=(InByte & 0b11100000) >> 5; // bit 5,6,7 are the command bits
					gbl_cur_param	=(InByte & 0b00011100) >> 2; // bit 2,3,4 are parameter bits for motors and sensors.
					gbl_cur_ext		=(InByte & 0b00000011);

					if (gbl_cur_cmd > ONE_BYTE_CMD)
					{	CMD_STATE = WAITING_FOR_SECOND_CMD_BYTE;
					} else
					{	CMD_STATE = CMD_READY;
					}
					break;
				case WAITING_FOR_SECOND_CMD_BYTE:

					 gbl_cur_ext_byte = InByte;
					 CMD_STATE = CMD_READY;
					 break;
				case CMD_READY:
					break;
				default:
		 			CMD_STATE = WAITING_FOR_FIRST_HEADER;
					break;
			}
			if (CMD_STATE == CMD_READY)  break;

	}

	if (buff_status == OVERFLOW)
	{	//Reset
		CMD_STATE = WAITING_FOR_FIRST_HEADER;

		//printf(ser_char,"Buffer overflow");
		//Halt();
	}
	// else assume no data in serial buffer
}



/////////////////////////////////////////////////////////////////////////
//
//   M A I N
//
/////////////////////////////////////////////////////////////////////////


void main() {

int current_cmd;
int cmd_status;
int Retries = RETRY;

int i=0;
char ch;
int *Port_PTR;

int buff_status;
int InByte;

init_variables();
initBoard();






while (1) {

	ProcessInput();


	if (CMD_STATE == CMD_READY)
	{

		if ((PASS_THROUGH == TRUE) && !((gbl_cur_cmd == CMD_PING) && (gbl_cur_param == 0)))
		{

			gbl_cur_cmd = (gbl_cur_cmd << 5) + (gbl_cur_param << 2) + gbl_cur_ext;
			//printf("%c",gbl_cur_cmd);
			printf(ser_char,"%c%c%c",ReplyHeader1, ReplyHeader2, ACK_BYTE);  // send acknowledge byte

		}
		else
		{
			switch (gbl_cur_cmd)  {
			case CMD_PING:
					Ping(gbl_cur_param);
					break;
			case CMD_READ_SENSOR:
					ReadSensor(gbl_cur_param, gbl_cur_ext);
					break;
			case CMD_MOTOR_CONTROL:

					MotorControl(gbl_cur_param);
					break;
			case CMD_MOTOR_POWER:
					SetMotorPower(gbl_cur_param);
					break;
			case CMD_TALK_TO_MOTOR:
					TalkToMotor(gbl_cur_ext_byte);
					break;
			case CMD_BURST_MODE:
					SetBurstMode(gbl_cur_ext_byte);
					break;
			}
	}
		CMD_STATE = WAITING_FOR_FIRST_HEADER;

		Retries = RETRY;
	}
	else if (CMD_STATE != WAITING_FOR_FIRST_HEADER)
	{

		Retries--;
		if (Retries <= 0)
		{	CMD_STATE = WAITING_FOR_FIRST_HEADER;
			Retries = RETRY;
		}
		delay_us(100);


//  One strange thing ... the code above didn't work when it was written like this
//		if (Retries-- < 0)
//		{	CMD_STATE = WAITING_FOR_FIRST_HEADER;
//			Retries = RETRY;
//			();
//		}
//		delay_us(10);

	}
	else   // NO_CMD
	{
		Retries = RETRY;
		DoSensorStuff();
	}

}

}





